use anyhow::{Context, Result};
use reqwest::StatusCode;
use serde::{Deserialize, Serialize};
use serde_json::json;
use sha2::{Sha256, Digest};
use std::collections::HashMap;
use uuid::Uuid;
use base64::{Engine as _, engine::general_purpose::STANDARD as BASE64};
use tracing::{info, error, debug, instrument};

use crate::vault_setup::{Department, User, VaultClient};

// Document status
#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum DocumentStatus {
    #[serde(rename = "pending")]
    Pending,
    #[serde(rename = "verified")]
    Verified,
}

// Document metadata
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Document {
    pub id: String,
    pub name: String,
    pub hash: String,
    pub status: DocumentStatus,
    pub signatures: HashMap<String, String>, // username -> signature
}

// Signature verification response
#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct SignatureVerification {
    pub document_id: String,
    pub is_verified: bool,
    pub signatures_count: usize,
    pub legal_signatures: usize,
    pub finance_signatures: usize,
    pub required_signatures: usize,
    pub required_legal: usize,
    pub required_finance: usize,
}

// Document service to handle document operations
pub struct DocumentService {
    vault_client: VaultClient,
}

impl DocumentService {
    pub fn new(vault_client: VaultClient) -> Self {
        DocumentService { vault_client }
    }

    // Upload a new document and store its metadata
    #[instrument(skip(self, content), fields(document_name = %name))]
    pub async fn upload_document(&self, name: &str, content: &[u8]) -> Result<String> {
        info!("Uploading new document: {}", name);

        // Generate a unique ID
        let id = Uuid::new_v4().to_string();
        debug!("Generated document ID: {}", id);

        // Calculate document hash
        let mut hasher = Sha256::new();
        hasher.update(content);
        let hash = format!("{:x}", hasher.finalize());
        debug!("Document hash: {}", hash);

        // Create document metadata
        let document = Document {
            id: id.clone(),
            name: name.to_string(),
            hash,
            status: DocumentStatus::Pending,
            signatures: HashMap::new(),
        };

        // Store document metadata in Vault
        self.store_document_metadata(&document).await?;

        info!("Document uploaded with ID: {}", id);
        Ok(id)
    }

    // Store document metadata in Vault
    #[instrument(skip(self, document), fields(document_id = %document.id))]
    async fn store_document_metadata(&self, document: &Document) -> Result<()> {
        debug!("Storing document metadata for {}", document.id);

        let url = format!("{}/v1/documents/data/docs/{}",
            self.vault_client.addr, document.id);

        let payload = json!({
            "data": {
                "id": document.id,
                "name": document.name,
                "hash": document.hash,
                "status": match document.status {
                    DocumentStatus::Pending => "pending",
                    DocumentStatus::Verified => "verified",
                },
                "signatures": document.signatures,
            }
        });

        let response = self.vault_client.client
            .post(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK | StatusCode::NO_CONTENT => {
                info!("Successfully stored document metadata for {}", document.id);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to store document metadata: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to store document metadata: {} - {}", status, error_text))
            }
        }
    }

    // Get document metadata
    #[instrument(skip(self))]
    pub async fn get_document(&self, document_id: &str) -> Result<Document> {
        debug!("Getting document metadata for {}", document_id);

        let url = format!("{}/v1/documents/data/docs/{}",
            self.vault_client.addr, document_id);

        let response = self.vault_client.client
            .get(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK => {
                let json: serde_json::Value = response.json().await?;

                // Extract status
                let status_str = json["data"]["data"]["status"]
                    .as_str()
                    .context("Failed to extract status")?;

                let status = match status_str {
                    "pending" => DocumentStatus::Pending,
                    "verified" => DocumentStatus::Verified,
                    _ => {
                        error!("Unknown document status: {}", status_str);
                        return Err(anyhow::anyhow!("Unknown status: {}", status_str));
                    }
                };

                // Extract signatures
                let signatures_value = &json["data"]["data"]["signatures"];
                let mut signatures = HashMap::new();

                if let Some(obj) = signatures_value.as_object() {
                    for (username, sig) in obj {
                        if let Some(sig_str) = sig.as_str() {
                            signatures.insert(username.clone(), sig_str.to_string());
                        }
                    }
                }

                let document = Document {
                    id: json["data"]["data"]["id"].as_str().context("Missing id")?.to_string(),
                    name: json["data"]["data"]["name"].as_str().context("Missing name")?.to_string(),
                    hash: json["data"]["data"]["hash"].as_str().context("Missing hash")?.to_string(),
                    status,
                    signatures,
                };

                debug!("Retrieved document: {} with {} signatures", document.id, document.signatures.len());
                Ok(document)
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to get document: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to get document: {} - {}", status, error_text))
            }
        }
    }

    // Sign a document with the user's key
    #[instrument(skip(self, user_token), fields(document_id = %document_id, username = %username))]
    pub async fn sign_document(&self, document_id: &str, username: &str, user_token: &str) -> Result<()> {
        info!("Signing document {} by user {}", document_id, username);

        // Get document metadata
        let document = self.get_document(document_id).await?;

        // Get user info to verify department
        let user = self.vault_client.get_user_info(username).await?;

        // Sign the document hash with user's key
        let url = format!("{}/v1/transit/sign/{}",
            self.vault_client.addr, username);

        let payload = json!({
            "input": BASE64.encode(document.hash.as_bytes()),
        });

        let response = self.vault_client.client
            .post(&url)
            .header("X-Vault-Token", user_token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK => {
                let json: serde_json::Value = response.json().await?;
                let signature = json["data"]["signature"]
                    .as_str()
                    .context("Failed to extract signature")?
                    .to_string();

                debug!("Generated signature for document {}", document_id);

                // Update document with signature
                self.add_signature(document_id, username, &signature).await?;

                // Update department signature record
                self.record_department_signature(document_id, &user).await?;

                // Check if document now has enough signatures
                self.update_document_status(document_id).await?;

                info!("Document {} signed by {}", document_id, username);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to sign document: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to sign document: {} - {}", status, error_text))
            }
        }
    }

    // Add a signature to a document
    #[instrument(skip(self, signature), fields(document_id = %document_id, username = %username))]
    async fn add_signature(&self, document_id: &str, username: &str, signature: &str) -> Result<()> {
        debug!("Adding signature from {} to document {}", username, document_id);

        // Get current document
        let mut document = self.get_document(document_id).await?;

        // Add signature
        document.signatures.insert(username.to_string(), signature.to_string());

        // Store updated document
        self.store_document_metadata(&document).await?;

        debug!("Added signature from {} to document {}", username, document_id);
        Ok(())
    }

    // Record department signature
    #[instrument(skip(self), fields(document_id = %document_id, department = ?user.department, username = %user.username))]
    async fn record_department_signature(&self, document_id: &str, user: &User) -> Result<()> {
        let dept_str = match user.department {
            Department::Legal => "legal",
            Department::Finance => "finance",
        };

        debug!("Recording {} department signature for document {}", dept_str, document_id);

        let url = format!("{}/v1/documents/data/dept/{}/signatures/{}",
            self.vault_client.addr, dept_str, document_id);

        // Check if department signatures already exist
        let response = self.vault_client.client
            .get(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .send()
            .await;

        let mut signatures = Vec::new();

        // If record exists, get current signatures
        if let Ok(resp) = response {
            if resp.status() == StatusCode::OK {
                let json: serde_json::Value = resp.json().await?;
                if let Some(array) = json["data"]["data"]["signatures"].as_array() {
                    for sig in array {
                        if let Some(sig_str) = sig.as_str() {
                            signatures.push(sig_str.to_string());
                        }
                    }
                }
            }
        }

        // Add user to signatures if not already present
        if !signatures.contains(&user.username) {
            signatures.push(user.username.clone());
        }

        // Store updated signatures
        let payload = json!({
            "data": {
                "signatures": signatures,
            }
        });

        let response = self.vault_client.client
            .post(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .json(&payload)
            .send()
            .await?;

        match response.status() {
            StatusCode::OK | StatusCode::NO_CONTENT => {
                info!("Recorded signature for {} in {} department", user.username, dept_str);
                Ok(())
            }
            status => {
                let error_text = response.text().await?;
                error!("Failed to record department signature: {} - {}", status, error_text);
                Err(anyhow::anyhow!("Failed to record department signature: {} - {}", status, error_text))
            }
        }
    }

    // Update document status if it has enough signatures
    #[instrument(skip(self))]
    async fn update_document_status(&self, document_id: &str) -> Result<()> {
        debug!("Checking if document {} has enough signatures", document_id);

        // Verify signatures
        let verification = self.verify_document_signatures(document_id).await?;

        if verification.is_verified {
            // Get current document
            let mut document = self.get_document(document_id).await?;

            // Update status
            document.status = DocumentStatus::Verified;

            // Store updated document
            self.store_document_metadata(&document).await?;

            info!("Document {} marked as verified", document_id);
        } else {
            debug!(
                "Document {} not yet verified. Has {}/{} signatures ({} legal, {} finance)",
                document_id,
                verification.signatures_count,
                verification.required_signatures,
                verification.legal_signatures,
                verification.required_finance
            );
        }

        Ok(())
    }

    // Verify document signatures
    #[instrument(skip(self))]
    pub async fn verify_document_signatures(&self, document_id: &str) -> Result<SignatureVerification> {
        info!("Verifying signatures for document {}", document_id);

        // Get document
        let document = self.get_document(document_id).await?;

        // Get signing requirements
        let url = format!("{}/v1/documents/data/config/signing_requirements",
            self.vault_client.addr);

        let response = self.vault_client.client
            .get(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .send()
            .await?;

        let config = match response.status() {
            StatusCode::OK => response.json::<serde_json::Value>().await?,
            status => {
                let error_text = response.text().await?;
                error!("Failed to get signing requirements: {} - {}", status, error_text);
                return Err(anyhow::anyhow!("Failed to get signing requirements: {} - {}", status, error_text));
            }
        };

        // Get required signatures
        let required_signatures = config["data"]["data"]["total_required"]
            .as_u64()
            .context("Missing total_required")? as usize;

        // Get department requirements
        let mut required_legal = 0;
        let mut required_finance = 0;

        if let Some(departments) = config["data"]["data"]["departments"].as_array() {
            for dept in departments {
                let name = dept["name"].as_str().context("Missing department name")?;
                let required = dept["required"].as_u64().context("Missing required")? as usize;

                match name {
                    "legal" => required_legal = required,
                    "finance" => required_finance = required,
                    _ => (),
                }
            }
        }

        // Get department signatures
        let legal_signatures = self.get_department_signatures(document_id, "legal").await?;
        let finance_signatures = self.get_department_signatures(document_id, "finance").await?;

        // Check if requirements are met
        let total_signatures = document.signatures.len();
        let is_verified = total_signatures >= required_signatures &&
                          legal_signatures.len() >= required_legal &&
                          finance_signatures.len() >= required_finance;

        let verification = SignatureVerification {
            document_id: document_id.to_string(),
            is_verified,
            signatures_count: total_signatures,
            legal_signatures: legal_signatures.len(),
            finance_signatures: finance_signatures.len(),
            required_signatures,
            required_legal,
            required_finance,
        };

        info!(
            "Verification result for document {}: verified={}, signatures={}/{}, legal={}/{}, finance={}/{}",
            document_id,
            verification.is_verified,
            verification.signatures_count,
            verification.required_signatures,
            verification.legal_signatures,
            verification.required_legal,
            verification.finance_signatures,
            verification.required_finance
        );

        Ok(verification)
    }

    // Get department signatures for a document
    #[instrument(skip(self))]
    async fn get_department_signatures(&self, document_id: &str, department: &str) -> Result<Vec<String>> {
        debug!("Getting {} department signatures for document {}", department, document_id);

        let url = format!("{}/v1/documents/data/dept/{}/signatures/{}",
            self.vault_client.addr, department, document_id);

        let response = self.vault_client.client
            .get(&url)
            .header("X-Vault-Token", &self.vault_client.token)
            .send()
            .await;

        let mut signatures = Vec::new();

        // If record exists, get signatures
        if let Ok(resp) = response {
            if resp.status() == StatusCode::OK {
                let json: serde_json::Value = resp.json().await?;
                if let Some(array) = json["data"]["data"]["signatures"].as_array() {
                    for sig in array {
                        if let Some(sig_str) = sig.as_str() {
                            signatures.push(sig_str.to_string());
                        }
                    }
                }
            }
        }

        debug!("Found {} signatures for {} department on document {}",
               signatures.len(), department, document_id);

        Ok(signatures)
    }
}
